# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from sfsdk import log, containers, utils
import os
import os.path
import yaml
import requests
import shutil

settings = {}

workspace_dir = None
settings_path = None

def get_container_conf(image_name):
    
    image_settings = settings['images'][image_name]['container']

    container_run_conf = {
        'image': image_settings['image_name'],
        'name': image_settings['container_name'],
        'ports': {
            # RDP
            3389: ( '127.0.0.1', image_settings['external_rdp_port']),
            # HTTP
            image_settings['internal_http_port']: ('127.0.0.1', image_settings['external_http_port']),
            },
        'cap_add': [] if image_settings['allow_internet'] else [ 'NET_ADMIN' ],
        'environment': { 
            'USR_PWD': image_settings['rdp_password'],
            'ALLOW_INTERNET': 'true' if image_settings['allow_internet'] else 'false'
            }
    }

    container_run_conf['environment'].update(image_settings.get('environment', {}))

    return container_run_conf

def get_settings(image_name):
    # Generate the settings for a built container
    return [
        {
            'name': 'image_name',
            'default': image_name,
            'validate': lambda answer: "Invalid choice" if not answer else True
            },
        {
            'name': 'container_name',
            'default': '%s-inst' % (image_name),
            'validate': lambda answer: "Invalid choice" if not answer else True
            },
        {
            'name': 'rdp_password',
            'default': 'password',
            'validate': lambda answer: "Invalid choice" if not answer else True
            },
        {
            'name': 'allow_internet',
            'default': True,
            },
        {
            'name': 'external_rdp_port',
            'default': '3389',
            'validate': lambda answer: "Invalid choice" if not answer.isdigit() or not 0 < int(answer) < 65535 else True
            },
        {
            'name': 'external_http_port',
            'default': '8050',
            'validate': lambda answer: "Invalid choice" if not answer.isdigit() or not 0 < int(answer) < 65535 else True
            },
        {
            'name': 'internal_http_port',
            'default': '80',
            'validate': lambda answer: "Invalid choice" if not answer.isdigit() or not 0 < int(answer) < 65535 else True
            },
        {
            'name': 'environment',
            'default': {}
        }
    ]

def _check_existing_image_structure(build_dir):

    build_fs_dir = os.path.join(build_dir, 'fs/')
    build_dockerfile = os.path.join(build_dir, 'Dockerfile')

    return os.path.exists(build_fs_dir) or os.path.exists(build_dockerfile)

def _create_image_structure(image_name, from_dir):

    image_settings = settings['images'][image_name]

    # Create base folder 
    os.makedirs(os.path.join(workspace_dir, image_name), exist_ok=True) 

    build_dir = image_settings['build_dir']

    build_fs_dir = os.path.join(build_dir, 'fs/') 
    build_dockerfile = os.path.join(build_dir, 'Dockerfile')

    os.makedirs(build_fs_dir, exist_ok=True) 

    # Warn if import has been requested but the build dir is not empty
    # This actually copies only the Dockerfile, app, and fs dir
    if from_dir:

        if not os.path.isdir(from_dir) or not os.listdir(from_dir):
            log.warn('Directory %s can\'t be imported because missing or empty' % (from_dir))
        else:
            utils.merge_tree(os.path.join(from_dir, 'fs'), build_fs_dir)
            shutil.copy(os.path.join(from_dir, 'Dockerfile'), build_dockerfile)
            log.success('Build directory has been copied to %s' % (utils.prettypath(build_dir)))

    if not os.path.exists(os.path.join(build_dir, 'Dockerfile')):
        containers.generate_dockerfile(build_dir)

def remove_image(image_name):

    img_dir = settings['images'][image_name]['build_dir']

    if os.path.isdir(img_dir):
        log.warn("Image directory exists, but it hasn't been deleted. Delete it manually running\n    rm -rf %s" % (img_dir))

    try:
        containers.image_remove(image_name)
    except log.FatalMsg as e:
        pass

    try:
        del settings['images'][image_name]
    except KeyError:
        raise log.FatalMsg("Didn't find image %s in images.yml" %  image_name)


    save()

def add_image_settings(image_name, build_dir = None, from_dir = None, container_settings = {}):

    if not build_dir:
        build_dir = os.path.join(workspace_dir, image_name)

    # build_dir must be absolute
    build_dir = os.path.abspath(build_dir)

    if settings.get('images').get(image_name):
        raise log.FatalMsg('Image %s already exists. Remove it with img-rm and retry.' % image_name)

    settings['images'][image_name] = {
            'build_dir': build_dir,
            'container': container_settings
        }

    if _check_existing_image_structure(build_dir):
        log.warn('Build directory %s already contains the image structure, skipping its initialization.' % build_dir)
    else:
        _create_image_structure(image_name, from_dir)

    save()

def load(base_dir):
    """ Load or create workspace and settings file"""
    global settings, workspace_dir, settings_path

    workspace_dir = os.path.join(
            base_dir,
            'img'
    )
    settings_path = os.path.join(
        base_dir, 
        'images.yml'
    )

    try:
        with open(settings_path, 'r') as yaml_file:
            settings = yaml.safe_load(yaml_file)
    except OSError as e:
        os.makedirs(workspace_dir, exist_ok=True)
        settings = {
                'images' : {},
                }
        save()
    except yaml.YAMLError as e:
        raise log.FatalMsg("Settings file %s has broken YAML, delete it to recreate" % settings_path)

def save():
    """ Save settings file"""
    global settings

    try:
        with open(settings_path, 'w+') as outfile:
            yaml.dump(settings, outfile, default_flow_style=False)
    except OSError as e:
        log.debug(str(e))
        os.makedirs(workspace_dir, exist_ok=True)
    except yaml.YAMLError as e:
        raise log.FatalMsg("Error saving current settings" % settings_path)

def get_package_path():
    return os.path.dirname(os.path.realpath(__file__))

def resolve_images(names):

    if len(names) == 1 and names[0] == 'all':
        return list(settings.get('images', {}).keys())
    else:
        return names

def get_snapshot_name(name):

    while name in settings['images'].keys():

        try:
            img_name, version_raw = name.rsplit('-v', 1)
            version = int(version_raw) + 1
        except (IndexError, ValueError):
            img_name = name
            version = 1

        name = '%s-v%d' % (img_name, version)

    return name



# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
import glob
import shutil
import json
import tempfile
import base64
import copy
import uuid
from sfsdk import log, utils, authsettings
from ruamel import yaml
from pathlib import Path

base_status_messages = {
    'ok': '0',
    'solved': '0',
    'fixed': '0',
    'nok': '1',
    'not-solved': '1',
    'vulnerable': '1',
    'broken-functionality': '2',
    'broken-login': '2',
    'broken-webserver': '2',
    'broken': '2',
    'not-installed': '2'
}

slug_components = [
    'technology',
    'framework',
    'title'
]

base_dir = None
exrs = {}

def load(workspace_dir):

    global base_dir, exrs

    base_dir = os.path.join(
            workspace_dir,
            'srcs',
            authsettings.current_source
        )

    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    exrs[authsettings.current_source] = {}

    for exr_yml_path in glob.glob('%s/exr/*/exercise.yml' % base_dir):

        meta_name_by_folder = os.path.join(*exr_yml_path.split('/')[-3:-1])
        
        loaded_meta = load_exercise_meta(meta_name_by_folder)

        meta_name_by_metadata = get_standard_exercise_name(loaded_meta)

        if meta_name_by_metadata in exrs[authsettings.current_source].keys():
            raise log.FatalMsg("Local exercise %s has a duplicate name" % (meta_name_by_metadata))
            
        exrs[authsettings.current_source][meta_name_by_metadata] = loaded_meta

def get_standard_exercise_name(meta_data):

    slug = '_'.join(re.sub(
            "[^a-z0-9]+",
            "_",
            str(meta_data[c].lower())
        ) for c in slug_components)

    long_uuid = meta_data['uuid']

    if long_uuid[:6] in ('local-', 'sfsdk-'):
        shorter_uuid = long_uuid[:10]
    else:
        shorter_uuid = long_uuid[:4]

    return 'exr/%s-%s' % (
        slug,
        shorter_uuid
    )

def get_standard_flag_filename(technology, flag_data):

    flag_name = flag_data['title'].lower()

    slug = re.sub(
        "[^a-zA-Z0-9]+",
        "_",
    flag_name)

    return 'flag_%s' % (slug)

def ls_exercises():

    return sorted(get_exercises().keys())

def get_exercises():

    return exrs[authsettings.current_source]

def get_exercise_name_data_by_uuid(exercise_uuid):

    return next(((n, d) for (n, d) in get_exercises().items() if d['uuid'] == exercise_uuid), (None, None))


def get_exercise_by_name(name):

    exercise_data = get_exercises().get(name)

    if exercise_data:
        return exercise_data
    else:
        raise log.FatalMsg('The exercise %s does not exist locally' % name)

def remove_exercise(exercise):

    exr_data = get_exercise_by_name(exercise)
        
    current_folder = os.path.join(base_dir, exercise)

    tempfolder = tempfile.mkdtemp()

    log.warn("Moving %s to %s" % (current_folder, tempfolder))

    shutil.move(current_folder, tempfolder)

    return exr_data['uuid']

def load_exercise_meta(exercise):

    exercise_dir = os.path.join(base_dir, exercise)

    if not os.path.exists(exercise_dir):
        os.makedirs(exercise_dir)

    exercise_meta_path = os.path.join(exercise_dir, 'exercise.yml')

    exercise_yaml = {}

    try:

        with open(exercise_meta_path, 'r') as stream:

            try:
                exercise_yaml = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise log.FatalMsg(exc)

            with open(os.path.join(exercise_dir, 'description.txt.md'), 'r') as stream:
                exercise_yaml['description'] = utils.load_and_replace_media_with_b64(
                    exercise_dir,
                    stream.read()
                )

            with open(os.path.join(exercise_dir, 'solution.md'), 'r') as stream:
                exercise_yaml['solution']['text'] = utils.load_and_replace_media_with_b64(
                    exercise_dir,
                    stream.read()
                )

            # Dirty validation on multiple flags with the same name
            added_flag_names = []

            for i, flag_yaml in enumerate(exercise_yaml['flags']):

                flag_name = get_standard_flag_filename(exercise_yaml['technology'], flag_yaml)

                if flag_name in added_flag_names:
                    raise log.FatalMsg("Error, flag %s with the same name detected" % (flag_name))

                added_flag_names.append(flag_name)

                # Dirty validation here
                if len(flag_yaml['flagList']) > 1:
                    raise log.FatalMsg("Error, more than one flagList per flag")

                flaglist = flag_yaml['flagList'][0]

                flaglist_md_path = '%s.md' % (flag_name)

                with open(os.path.join(exercise_dir, flaglist_md_path), 'r') as stream:
                    exercise_yaml['flags'][i]['flagList'][0]['md']['text'] = utils.load_and_replace_media_with_b64(
                        exercise_dir,
                        stream.read()
                    )

                if 'hint' in flaglist and 'text' in flaglist['hint']['md']:
                    flaglist_hint_md_path = '%s_hint.md' % (flag_name)

                    with open(os.path.join(exercise_dir, flaglist_hint_md_path), 'r') as stream:
                        exercise_yaml['flags'][i]['flagList'][0]['hint']['md']['text'] = utils.load_and_replace_media_with_b64(
                            exercise_dir,
                            stream.read()
                        )

                if exercise_yaml.get('labType') in ( '1', 'CODE_REVIEW'):

                    with open(os.path.join(exercise_dir, '%s_lines.yml' % (flag_name)), "r") as f:

                        try:
                            crlines = yaml.safe_load(f)
                        except yaml.YAMLError as exc:
                            raise log.FatalMsg(exc)

                        exercise_yaml['flags'][i]['flagList'][0]['crVulns'] =  crlines.get('crVulns', [])

                if flaglist.get('selfCheckAvailable'):
                    
                    with open(os.path.join(exercise_dir, '%s_messages.yml' % (flag_name)), "r") as f:

                        try:
                            messages_yaml = yaml.safe_load(f)
                        except yaml.YAMLError as exc:
                            raise log.FatalMsg(exc)

                        exercise_yaml['flags'][i]['flagList'][0]['selfCheck']['messageMappings'] = {}
                        exercise_yaml['flags'][i]['flagList'][0]['selfCheck']['statusMapping'] = {}

                        if messages_yaml and type(messages_yaml) is dict:
                            for status_name, status_message in messages_yaml.items():

                                # Do not add if loaded status name is a base status message
                                if status_name in base_status_messages:
                                    continue

                                # Add only if loaded status name starts with a base status message
                                base_status_number = next((v for m,v in base_status_messages.items() if status_name.startswith(m + '-')), None)
                                if base_status_number == None:
                                    raise log.FatalMsg("Can't map message status '%s' on a basic status type" % (status_name))

                                exercise_yaml['flags'][i]['flagList'][0]['selfCheck']['messageMappings'][status_name] = status_message
                                exercise_yaml['flags'][i]['flagList'][0]['selfCheck']['statusMapping'][status_name] = base_status_number
                        
    except FileNotFoundError as e:
        raise log.FatalMsg(f"Can't load the metadata: {e.filename} was expected but not found.")

    return exercise_yaml

def save_exercise_meta(exercise_yaml_orig):

    exercise_yaml = copy.deepcopy(exercise_yaml_orig)

    exercise = get_standard_exercise_name(exercise_yaml)

    exercise_dir = os.path.join(base_dir, exercise)

    if not os.path.exists(exercise_dir):
        os.makedirs(exercise_dir)

    if log.verbose:
        exercise_meta_debug_path = os.path.join(exercise_dir, '.last-saved.json')
        log.debug('Saving JSON debug to %s' % (exercise_meta_debug_path))
        with open(exercise_meta_debug_path, 'w') as f:
              json.dump(exercise_yaml, f, indent=4)

    description = exercise_yaml['description']
    description = utils.save_and_replace_b64_with_media(
        base_folder = exercise_dir,
        content = description,
        content_name = 'description'
    )
    with open(os.path.join(exercise_dir, 'description.txt.md'), 'w+') as stream:
        stream.write(description)

    exercise_yaml['description'] = '!import description.txt.md'

    solution_text = exercise_yaml['solution']['text']
    solution_text = utils.save_and_replace_b64_with_media(
        base_folder = exercise_dir,
        content = solution_text,
        content_name = 'solution'
    )
    with open(os.path.join(exercise_dir, 'solution.md'), 'w+') as stream:
        stream.write(solution_text)

    exercise_yaml['solution']['text'] = '!import solution.md'

    # Do not store the complete stack KB, but only variant and technology
    try:
        exercise_yaml['stack'] = {
            'technology' : exercise_yaml['stack']['technology'],
            'variant': exercise_yaml['stack']['variant'],
            'uuid': exercise_yaml['stack']['uuid']
        }
    except KeyError as e:
        raise log.FatalMsg(f"Metadata parsing error {e}")


    # Dirty validation on multiple flags with the same name
    added_flag_names = []

    for i, flag_yaml in enumerate(exercise_yaml['flags']):

        flag_name = get_standard_flag_filename(exercise_yaml['technology'], flag_yaml)

        if flag_name in added_flag_names:
            raise log.FatalMsg("Error, flag %s with the same name detected" % (flag_name))
        added_flag_names.append(flag_name)

        try:
            # Do not store the complete vulnerability KB, but only category, vulnerability, technology, and isAgnostic
            exercise_yaml['flags'][i]['kb'] = {
                'category' : exercise_yaml['flags'][i]['kb']['category'],
                'vulnerability': exercise_yaml['flags'][i]['kb']['vulnerability'],
                'technology': exercise_yaml['flags'][i]['kb']['technology'],
                'isAgnostic': exercise_yaml['flags'][i]['kb']['isAgnostic'],
                'uuid': exercise_yaml['flags'][i]['kb']['uuid']
            }
        except KeyError as e:
            raise log.FatalMsg(f"Metadata parsing error {e}")

        # Dirty validation on multiple flagList
        if len(flag_yaml['flagList']) > 1:
            raise log.FatalMsg("Error, more than one flagList per flag")

        flaglist = flag_yaml['flagList'][0]
            
        flaglist_md = flaglist['md']['text']
        flaglist_md_path = '%s.md' % (flag_name)

        flaglist_md = utils.save_and_replace_b64_with_media(
            base_folder = exercise_dir,
            content = flaglist_md,
            content_name = flag_name
        )

        with open(os.path.join(exercise_dir, flaglist_md_path), 'w+') as stream:
            stream.write(flaglist_md)

        exercise_yaml['flags'][i]['flagList'][0]['md']['text'] = '!import %s' % flaglist_md_path

        if 'hint' in flaglist and 'text' in flaglist['hint']['md']:
            flaglist_hint_md = flaglist['hint']['md']['text']
            flaglist_hint_md_path = '%s_hint.md' % (flag_name)

            flaglist_hint_md = utils.save_and_replace_b64_with_media(
                base_folder = exercise_dir,
                content = flaglist_hint_md,
                content_name = f'{flag_name}_hint'
            )

            with open(os.path.join(exercise_dir, flaglist_hint_md_path), 'w+') as stream:
                stream.write(flaglist_hint_md)

            exercise_yaml['flags'][i]['flagList'][0]['hint']['md']['text'] = '!import %s' % flaglist_hint_md_path

        if flaglist.get('selfCheckAvailable'):
            with open(os.path.join(exercise_dir, "%s_messages.yml" % (flag_name)), "w") as f:
                yaml.dump(flaglist['selfCheck']['messageMappings'], f, default_flow_style=False, width=float("inf"))
                exercise_yaml['flags'][i]['flagList'][0]['selfCheck']['messageMappings'] = "!import %s_messages.yml" % (flag_name)
        
        if exercise_yaml.get('labType') in ( '1', 'CODE_REVIEW'):

            with open(os.path.join(exercise_dir, '%s_lines.yml' % (flag_name)), "w+") as f:

                yaml.dump({
                    'crVulns': flaglist['crVulns'],
                }, f, default_flow_style=False, width=float("inf"))

                exercise_yaml['flags'][i]['flagList'][0]['crVulns'] = '!import %s_lines.yml crVulns' % (flag_name)

    fix_incoming_exercise_data(exercise_yaml)

    with open(os.path.join(exercise_dir, "exercise.yml"), "w") as f:
        yaml.dump(exercise_yaml, f, default_flow_style=False, width=float("inf"))

    return exercise_dir

def copy_exercise_folder(old_exercise_dir, new_exercise_data):

    new_exercise_dir = save_exercise_meta(new_exercise_data)
    log.warn("Copied from %s to %s" % (utils.prettypath(old_exercise_dir), utils.prettypath(new_exercise_dir)))

def update_exercise_folder(old_exercise_dir, new_exercise_data):

    new_exercise_dir = save_exercise_meta(new_exercise_data)

    if old_exercise_dir != new_exercise_dir:
        utils.trash_file(old_exercise_dir)
        log.warn("Moved from %s to %s" % (utils.prettypath(old_exercise_dir), utils.prettypath(new_exercise_dir)))


def get_exercise_environment_variable(exercise_name):

    exercise_settings = get_exercise_by_name(exercise_name)
    return base64.b64encode(json.dumps(
            {
                'title': exercise_settings['title'],
                'flags': exercise_settings['flags']
            }
        ).encode()
    )

def generate_exercise_environment(flag_name):

    return base64.b64encode(json.dumps(
            {
                'flags': [{"flagList":[{"selfCheck":{"name": flag_name }}]}]
            }
        ).encode()
    )

def fix_incoming_exercise_data(exercise_json):

    # Int-ify status
    exercise_json['status'] = int(exercise_json['status'])

    # Sort by flag type, to keep always the same order
    exercise_json['flags'].sort(key=lambda f: f['flagList'][0]['type'])

    # Delete ids
    utils.recursive_del(exercise_json['flags'], 'id')
    
    try:
        del exercise_json['solution']['id']
    except KeyError:
        pass

    # Map labType to strings
    labType = exercise_json.get('labType')
    if not labType or labType == '0':
        exercise_json['labType'] = 'LAB'
    elif labType == '1':
        exercise_json['labType'] = 'CODE_REVIEW'
    elif labType == '2':
        exercise_json['labType'] = 'QUIZ'

def fix_outgoing_exercise_data(exercise_json):

    # Int-ify status
    exercise_json['status'] = int(exercise_json['status'])

    # Delete any reference to KBs and Stack UUIDs, let the platform guess by other meta
    utils.recursive_del(exercise_json['flags'], 'uuid')
    utils.recursive_del(exercise_json['stack'], 'uuid')

    # Delete md.id in flags that cause some issue
    utils.recursive_del(exercise_json['flags'], 'id')

    # Map labType to strings
    labType = exercise_json.get('labType')
    if not labType or labType == '0':
        exercise_json['labType'] = 'LAB'
    elif labType == '1':
        exercise_json['labType'] = 'CODE_REVIEW'
    elif labType == '2':
        exercise_json['labType'] = 'QUIZ'

def get_path(exercise_name):
   return os.path.join(base_dir, exercise_name)

def get_pretty_path(exercise_name):
   return utils.prettypath(os.path.join(base_dir, exercise_name))


# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from sfsdk import exrsettings, log, utils, authsettings
from ruamel import yaml
import os
import shutil
import glob
import re
import json
import copy
import uuid
import tempfile

base_dir = None

stacks = {}

slug_components = [
    'technology',
    'variant'
]

def load(workspace_dir):

    global base_dir, exrs

    base_dir = os.path.join(
            workspace_dir,
            'srcs',
            authsettings.current_source
        )

    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    stacks[authsettings.current_source] = {}

    loaded_uuids = {} # To check duplicates
    for stack_yml_path in glob.glob('%s/stack/*/stack.yml' % base_dir):

        meta_name_by_folder = os.path.join(*stack_yml_path.split('/')[-3:-1])
        
        loaded_meta = load_stack_meta(meta_name_by_folder)

        meta_name_by_metadata = get_standard_stack_name(loaded_meta)

        # Do not consider meta-changed exercises as wrong
        if loaded_meta['uuid'] != 'xxxx' and meta_name_by_metadata != meta_name_by_folder:
            raise log.FatalMsg("Stack %s in incorrect folder %s" % (meta_name_by_metadata, stack_yml_path))

        if meta_name_by_metadata in stacks[authsettings.current_source].keys():
            raise log.FatalMsg("Local stack %s has a duplicate name" % (meta_name_by_metadata))
            
        if loaded_meta['uuid'] != 'xxxx' and loaded_meta['uuid'] in loaded_uuids:
            raise log.FatalMsg("Local stacks %s and %s have the same uuid" % (meta_name_by_metadata, loaded_uuids[loaded_meta['uuid']]))

        stacks[authsettings.current_source][meta_name_by_metadata] = loaded_meta
        loaded_uuids[loaded_meta['uuid']] = meta_name_by_metadata

def get_standard_stack_name(meta_data):

    slug = '_'.join(re.sub(
            "[^a-z0-9]",
            "_",
            str(meta_data[c].lower())
        ) for c in slug_components)

    long_uuid = meta_data['uuid']

    if long_uuid[:6] in ('local-', 'sfsdk-'):
        shorter_uuid = long_uuid[:10]
    else:
        shorter_uuid = long_uuid[:4]

    return 'stack/%s-%s' % (
        slug,
        shorter_uuid
    )

def save_stack_meta(stack_yaml_orig):

    stack_yaml = copy.deepcopy(stack_yaml_orig)

    stack = get_standard_stack_name(stack_yaml)

    stack_dir = os.path.join(base_dir, stack)

    if not os.path.exists(stack_dir):
        os.makedirs(stack_dir)

    information_text = stack_yaml['md']['text']
    information_text = utils.save_and_replace_b64_with_media(
        base_folder = stack_dir,
        content = information_text,
        content_name = 'stack'
    )
    with open(os.path.join(stack_dir, 'stack.md'), 'w+') as f:
        f.write(information_text)
    stack_yaml['md']['text'] = '!import stack.md'

    reload_text = stack_yaml['mdReload'].get('text')
    if reload_text:

        reload_text = utils.save_and_replace_b64_with_media(
            base_folder = stack_dir,
            content = reload_text,
            content_name = 'reload'
        )
        with open(os.path.join(stack_dir, 'reload.md'), 'w+') as f:
            f.write(reload_text)
        stack_yaml['mdReload']['text'] = '!import reload.md'

    if log.verbose:
        stack_meta_debug_path = os.path.join(stack_dir, '.last-saved.json')
        log.debug('Saving JSON debug to %s' % (stack_meta_debug_path))
        with open(stack_meta_debug_path, 'w') as f:
              json.dump(stack_yaml_orig, f, indent=4)

    with open(os.path.join(stack_dir, "stack.yml"), "w") as f:
        yaml.dump(stack_yaml, f, default_flow_style=False)

    return stack_dir

def load_stack_meta(stack):

    stack_dir = os.path.join(base_dir, stack)

    stack_meta_path = os.path.join(stack_dir, 'stack.yml')

    with open(stack_meta_path, 'r') as stream:
        try:
            stack_yaml = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            raise log.FatalMsg(exc)

        with open(os.path.join(stack_dir, 'stack.md'), 'r') as stream:
            stack_yaml['md']['text'] = utils.load_and_replace_media_with_b64(
                stack_dir,
                stream.read()
            )

        reload_md_path = os.path.join(stack_dir, 'reload.md')
        if os.path.exists(reload_md_path):
            with open(reload_md_path, 'r') as stream:
                stack_yaml['mdReload']['text'] = utils.load_and_replace_media_with_b64(
                    stack_dir,
                    stream.read()
                )
            stack_yaml['restart'] = True

    return stack_yaml

def get_stacks():

    return stacks[authsettings.current_source]

def get_stack_name_data_by_uuid(stack_uuid):

    return next(((n, d) for (n, d) in get_stacks().items() if d['uuid'] == stack_uuid), (None, None))

def copy_stack_folder(old_stack_dir, new_stack_data):
    new_stack_dir = save_stack_meta(new_stack_data)
    log.warn("Copied from %s to %s" % (utils.prettypath(old_stack_dir), utils.prettypath(new_stack_dir)))

def update_stack_folder(old_stack_dir, new_stack_data):

    new_stack_dir = save_stack_meta(new_stack_data)

    if old_stack_dir != new_stack_dir:
        utils.trash_file(old_stack_dir)
        log.warn("Moved from %s to %s" % (utils.prettypath(old_stack_dir), utils.prettypath(new_stack_dir)))

def delete_stack_by_uuid(stack_uuid):

    current_name, stack_data = get_stack_name_data_by_uuid(stack_uuid)
        
    current_folder = os.path.join(base_dir, current_name)

    if stack_data['uuid'] == stack_uuid:

        tempfolder = tempfile.mkdtemp()

        print("Moving %s to %s" % (current_folder, tempfolder))

        shutil.move(current_folder, tempfolder)

        return stack_uuid


# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from sfsdk import exrsettings, log, utils, authsettings
from ruamel import yaml
import os
import shutil
import glob
import re
import shutil
import json
import copy
import tempfile

base_dir = None

kbs = {}

slug_components = [
    'technology',
    'vulnerability'
]

def load(workspace_dir):

    global base_dir, exrs

    base_dir = os.path.join(
            workspace_dir,
            'srcs',
            authsettings.current_source
        )

    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    kbs[authsettings.current_source] = {}

    for kb_yml_path in glob.glob('%s/kb/*/kb.yml' % base_dir):

        meta_name_by_folder = os.path.join(*kb_yml_path.split('/')[-3:-1])
        
        loaded_meta = load_kb_meta(meta_name_by_folder)

        meta_name_by_metadata = get_standard_kb_name(loaded_meta)

        if meta_name_by_metadata in kbs[authsettings.current_source].keys():
            raise log.FatalMsg("Local kb %s has a duplicate name" % (meta_name_by_metadata))
            
        kbs[authsettings.current_source][meta_name_by_metadata] = loaded_meta

def get_standard_kb_name(meta_data):

    slug = '_'.join(re.sub(
            "[^a-z0-9]",
            "_",
            str(meta_data[c].lower())
        ) for c in slug_components)

    long_uuid = meta_data['uuid']

    if long_uuid[:6] in ('local-', 'sfsdk-'):
        shorter_uuid = long_uuid[:10]
    else:
        shorter_uuid = long_uuid[:4]

    return 'kb/%s-%s' % (
        slug,
        shorter_uuid
    )

def save_kb_meta(kb_yaml_orig):

    kb_yaml = copy.deepcopy(kb_yaml_orig)

    kb = get_standard_kb_name(kb_yaml)

    kb_dir = os.path.join(base_dir, kb)

    if not os.path.exists(kb_dir):
        os.makedirs(kb_dir)


    information_text = kb_yaml['md']['text']
    information_text = utils.save_and_replace_b64_with_media(
        base_folder = kb_dir,
        content = information_text,
        content_name = 'kb'
    )
    with open(os.path.join(kb_dir, 'kb.md'), 'w+') as f:
        f.write(information_text)
    kb_yaml['md']['text'] = '!import kb.md'

    if log.verbose:
        kb_meta_debug_path = os.path.join(kb_dir, '.last-saved.json')
        log.debug('Saving JSON debug to %s' % (kb_meta_debug_path))
        with open(kb_meta_debug_path, 'w') as f:
              json.dump(kb_yaml_orig, f, indent=4)

    with open(os.path.join(kb_dir, "kb.yml"), "w") as f:
        yaml.dump(kb_yaml, f, default_flow_style=False)

    return kb_dir

def load_kb_meta(kb):

    kb_dir = os.path.join(base_dir, kb)

    if not os.path.exists(kb_dir):
        os.makedirs(kb_dir)

    kb_meta_path = os.path.join(kb_dir, 'kb.yml')

    kb_yaml = {}

    try:

        with open(kb_meta_path, 'r') as stream:
            try:
                kb_yaml = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise log.FatalMsg(exc)

            with open(os.path.join(kb_dir, 'kb.md'), 'r') as stream:
                kb_yaml['md']['text'] = utils.load_and_replace_media_with_b64(
                        kb_dir,
                        stream.read()
                    )

            

    except IOError:
        pass

    return kb_yaml

def get_kbs():

    return kbs[authsettings.current_source]

def get_kb_name_data_by_uuid(kb_uuid):
    
    return next(((n, d) for (n, d) in get_kbs().items() if d['uuid'] == kb_uuid), (None, None))


def delete_kb_by_uuid(kb_uuid):

    current_name, kb_data = get_kb_name_data_by_uuid(kb_uuid)
        
    current_folder = os.path.join(base_dir, current_name)

    if kb_data['uuid'] == kb_uuid:

        tempfolder = tempfile.mkdtemp()

        log.warn("Moving %s to %s" % (current_folder, tempfolder))

        shutil.move(current_folder, tempfolder)

        return kb_uuid

def copy_kb_folder(old_kb_dir, new_kb_data):
    new_kb_dir = save_kb_meta(new_kb_data)
    log.warn("Copied from %s to %s" % (utils.prettypath(old_kb_dir), utils.prettypath(new_kb_dir)))

def update_kb_folder(old_kb_dir, new_kb_data):

    new_kb_dir = save_kb_meta(new_kb_data)

    if old_kb_dir != new_kb_dir:
        utils.trash_file(old_kb_dir)
        log.warn("Moved from %s to %s" % (utils.prettypath(old_kb_dir), utils.prettypath(new_kb_dir)))



# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import textwrap
import logging
import prettytable

class clrs:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[33m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

verbose = False
quiet = False

class FatalMsg(Exception):
    pass

def set_verbose():
    global verbose

    verbose = True

    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("urllib3").setLevel(logging.WARNING) # Disable urllib3 debug

def set_quiet():
    global quiet

    quiet = True

# Warning
def warn(msg):
    if not quiet: print(f'[{clrs.WARNING}!{clrs.ENDC}] {msg}')

# Success
def success(msg):
    if not quiet: print(f'[{clrs.OKGREEN}+{clrs.ENDC}] {msg}')

# Info
def info(msg):
    if not quiet: print(f'[{clrs.OKGREEN}*{clrs.ENDC}] {msg}')

# Multiline prints
def block(msg, end = '', prefix = ''):
   if not quiet: print(textwrap.indent(textwrap.dedent(msg), prefix = prefix), end = end)

def debug(msg):
    if not quiet and verbose: logging.debug(msg)

def tablify(data, header = []):

    if quiet:
        return

    table = prettytable.PrettyTable()

    for row in data:
        table.add_row(row)

    table.align = 'l'
    table.vrules = prettytable.NONE
    table.hrules = prettytable.HEADER
    table.left_padding_width = 1

    if header:
        table.field_names = header
    else:
        table.header = False

    print('\n' + str(table))




# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import stat
import re
from sfsdk import log
import os
import base64
import mimetypes
import tempfile

re_extract_b64_data = re.compile(r'data:([a-zA-Z0-9/]+);base64,([A-Za-z0-9+/=]+)')
re_extract_src_media = re.compile(r'src\=(?:\"|\')(.+?)(?:\"|\')')
re_extract_markdown_media = re.compile(r'!\[.*?\]\((.*?)\)')

def load_and_replace_media_with_b64(base_folder, content):

    content_with_no_code_blocks = re.sub(r'`.*?`', '', re.sub(r'```.*?```', '', content, flags=re.DOTALL))

    for media_path in re_extract_src_media.findall(content_with_no_code_blocks) + re_extract_markdown_media.findall(content_with_no_code_blocks):

        if not os.path.isfile(media_path):

            absolute_media_path = os.path.join(base_folder, media_path)

            if not os.path.exists(absolute_media_path):
                continue
            
            confirmed_media_path = absolute_media_path

        else:
            confirmed_media_path = media_path

        mime_type = mimetypes.guess_type(confirmed_media_path)[0]
        
        with open(confirmed_media_path, 'rb') as stream:
            media_content = stream.read()
            b64_data = base64.b64encode(media_content).decode()

        b64_string = f'data:{mime_type};base64,{b64_data}'

        content = content.replace(media_path, b64_string)

    return content
            

def save_and_replace_b64_with_media(base_folder, content, content_name):

    for idx, b64_mime_data in enumerate(re_extract_b64_data.findall(content)):

        mime_type, b64_data = b64_mime_data

        extension = mimetypes.guess_extension(mime_type)

        file_name = f'{content_name}_{idx}{extension}'

        media_relative_path = os.path.join(
            'media',
            file_name
        )
        os.makedirs(
            os.path.join(base_folder, 'media'), 
            exist_ok = True
        )

        media_absolute_path = os.path.join(
            base_folder, media_relative_path
        )

        with open(media_absolute_path, 'wb+') as stream:
            stream.write(base64.b64decode(b64_data))

        content = content.replace(
            f'data:{mime_type};base64,{b64_data}', 
            media_relative_path
        )

        log.debug(f'Saved {extension} media file {media_relative_path}')
    
    return content

def merge_tree(src, dst, preserve_symlinks = True, ignore = None):
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    if ignore:
        excl = ignore(src, lst)
        lst = [x for x in lst if x not in excl]
    for item in lst:
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if preserve_symlinks and os.path.islink(s):
             if os.path.lexists(d):
                 os.remove(d)
             os.symlink(os.readlink(s), d)
             try:
                 st = os.lstat(s)
                 mode = stat.S_IMODE(st.st_mode)
                 os.lchmod(d, mode)
             except:
                 pass # lchmod not available
        elif os.path.isdir(s):
            merge_tree(s, d, preserve_symlinks, ignore)
        else:
            shutil.copy2(s, d)


def recursive_del(obj, bad_key):
    if isinstance(obj, dict):
        # the call to `list` is useless for py2 but makes
        # the code py2/py3 compatible
        for key in list(obj.keys()):
            if key == bad_key:
                del obj[key]
            else:
                recursive_del(obj[key], bad_key)
    elif isinstance(obj, list):
        for i in reversed(range(len(obj))):
            recursive_del(obj[i], bad_key)
    else:
        pass

def prettypath(p):
    return p.replace(os.path.expanduser('~'), '~', 1)


def trash_file(file_path, suffix = 'file'):

    tmpdirname = tempfile.mkdtemp(suffix = '-trashed-sf-%s' % (suffix))
    shutil.move(file_path, tmpdirname)
    
    log.debug('Trashed %s moving it to %s' % (file_path, tmpdirname))


# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from sfsdk import log
from sfsdk import authsettings
import boto3
import requests
import time

r = requests.Session()

remote_archive_format = '%s-%s-%d.zip'

hub_url = None
hub_token = None
role = None

def login_to_hub(username, password, management_hub_url):

    global hub_token, hub_url

    hub_url = management_hub_url
    
    source = authsettings.current_source
    r.headers.update(
            authsettings.settings['sources'][source].get('headers', {})
        )

    res = r.post(
            hub_url + '/rest/api/login-developer',
            json = {
                "username": username,
                "password": password
                } 
            )

    hub_repr = 'SecureFlag' if management_hub_url == authsettings.default_hub_url else source

    if res.status_code != 200:
        raise log.FatalMsg("Failed authentication to the %s hub as user %s, check credentials in %s" % (hub_repr, username, authsettings.settings_path))
    else:
        log.success("Logged in the %s hub as user %s" % (hub_repr, username))
        hub_token = res.json()

def get_s3_creds_via_hub():

    res = r.get(
            hub_url + '/rest/user/uploads/credentials',
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            }
        )

    return res.json()

def get_arn_via_hub():

    res = r.get(
            hub_url + '/rest/user/uploads/bucket',
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            }
        )

    return res.json()['arn']
    
def get_frameworks_via_hub():

    res = r.get(
            hub_url + '/rest/user/frameworks',
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            })
    
    if res.status_code == 204:
        return []

    return res.json()

def get_technologies_via_hub():

    res = r.get(
            hub_url + '/rest/user/technologies',
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            }
        )
    
    if res.status_code == 204:
        return []

    return res.json()

def get_technology_via_hub_by_uuid(uuid):

    res = r.get(
            hub_url + '/rest/user/technologies/%s' % uuid,
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            }
        )

    if res.status_code == 204:
        return {}

    return res.json()

def get_vulnerabilities_via_hub():

    res = r.get(
            hub_url + '/rest/user/vulnerabilities',
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            }
        )

    if res.status_code == 204:
        return []

    return res.json()

def get_vulnerability_via_hub_by_uuid(uuid):

    res = r.get(
            hub_url + '/rest/user/vulnerabilities/%s' % uuid,
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            }
        )

    if res.status_code == 204:
        return {}

    return res.json()

def get_exercises_via_hub():

    res = r.get(
            hub_url + '/rest/user/exercises',
            headers = {
                'Authorization': 'Bearer %s' % (hub_token['value'])
            }
        )

    if res.status_code == 204:
        return {}

    return res.json()

def push_to_s3(archive_path, aws_creds, username, aws_arn):

    stsclient = boto3.client(
        'sts',
        aws_access_key_id=aws_creds['accessKeyId'],
        aws_secret_access_key=aws_creds['secretAccessKey'],
        aws_session_token=aws_creds['sessionToken'],
    )
    account_identity = stsclient.get_caller_identity()

    s3client = boto3.client(
        's3',
        aws_access_key_id=aws_creds['accessKeyId'],
        aws_secret_access_key=aws_creds['secretAccessKey'],
        aws_session_token=aws_creds['sessionToken'],
    )

    remote_archive_name = remote_archive_format % (
        account_identity['Account'],
        username,
        int(time.time())
    )

    bucket = aws_arn.split(":")[5]

    try:
        res = s3client.upload_file(
            archive_path + '.zip', 
            bucket, 
            remote_archive_name
        )
    except boto3.exceptions.S3UploadFailedError as e:
        raise log.FatalMsg(str(e))

    log.success('Exercise has been published as %s' % (remote_archive_name))


#!/usr/bin/env python3

# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import argparse
import json
import os
import re
import shlex
import shutil
import sys
import tempfile
import time

from sfsdk import (authsettings, containers, exrsettings, frameworksettings,
                   imgsettings, kbsettings, log, remote, stacksettings,
                   utils, web)

base_dir = os.path.expanduser('~/sf')

def img_snapshot(name, new_name, force):

    image_settings = imgsettings.settings.get('images', {}).get(name)

    if not new_name:
        new_name = imgsettings.get_snapshot_name(name)

    if not image_settings:
        raise log.FatalMsg("Image %s does not exist, run 'sfsdk img-add' to add it" % name)

    if new_name in imgsettings.settings.get('images', {}).keys():
        raise log.FatalMsg('Image %s already exists' % new_name)

    current_container_conf = imgsettings.get_container_conf(name)

    if containers.container_status(current_container_conf['name']) != 'running':
        raise log.FatalMsg("""Image is not running, consider running 'sfsdk img-watch %s' before running img-snapshot again.""" % (name))

    additional_paths = []

    paths_file = os.path.join(image_settings['build_dir'], '.sfsdk-watch.txt')

    additional_paths = []
    try:
        with open(paths_file, 'r') as stream:
            additional_paths = [ l for l in stream.read().split('\n') if l ]
    except OSError:
        log.debug("Error reading %s" % (paths_file))
        log.warn("Can't find result from img-watch, consider running 'sfsdk img-watch %s' before running img-snapshot again." % (name))

    monitored_paths = []

    # Add additional paths without duplicates
    for additional_path in additional_paths:
        if not (additional_path in monitored_paths or additional_path + '/' in monitored_paths):
            monitored_paths.append(additional_path)

    container_settings = { v['name']:v['default'] for v in imgsettings.get_settings(new_name) }

    imgsettings.add_image_settings(
            new_name,
            build_dir = None,
            from_dir = image_settings['build_dir'],
            container_settings = container_settings
            )

    new_img_settings = imgsettings.settings.get('images', {}).get(new_name)
    new_build_dir = new_img_settings['build_dir']

    containers.save_container_snapshot(
            container_name = current_container_conf['name'],
            snapshot_folder = new_build_dir,
            modified_paths = monitored_paths
        )

    if monitored_paths:
        log.block("""The following paths have been exported from the running container and applied over the build folder.:

%s

The new project image has been saved as %s.
""" % (
            '\n'.join(monitored_paths),
            new_name
            )
        )

def img_add(name, build_dir, from_dir):

    # Docker image name limitations
    if not re.match('[a-z0-9][a-z0-9_.-]*$', name):
        raise log.FatalMsg('Image name is not valid')

    container_settings = { v['name']:v['default'] for v in imgsettings.get_settings(name) }

    imgsettings.add_image_settings(
            name,
            build_dir = build_dir,
            from_dir = from_dir,
            container_settings = container_settings
            )

    image_settings = imgsettings.settings.get('images', {}).get(name)

    log.success("""New image %s has been added at '%s'""" % (name, utils.prettypath(image_settings['build_dir'])))

def img_build(name, source):

    image_settings = imgsettings.settings.get('images', {}).get(name)

    if not image_settings:
        raise log.FatalMsg("Image %s does not exist, run 'sfsdk img-add %s' to add it" % (name, name))

    log.info('Building exercise image %s...' % name)
    build_generator = containers.build_image(image_settings['build_dir'], name)

    built = False
    error = None

    for line in build_generator:
        if 'stream' in line:
            output = line['stream'].strip('\n')
            log.block(output, end='\n', prefix='')

            if output.startswith('Successfully built'):
                built = True

        if 'error' in line:
            error = line['error']

    if built:
        log.success("Build complete")
    else:
        raise log.FatalMsg("The build of the image failed: %s" % error)

def img_shell(name, command):

    image_settings = imgsettings.settings.get('images', {}).get(name)

    if not image_settings:
        raise log.FatalMsg("Image %s does not exist, run 'sfsdk img-add' to add it" % name)

    container_settings = imgsettings.get_container_conf(name)
    container_name = container_settings['name']

    os.execvp("docker", [ "docker", "exec", "-it", container_name ] + shlex.split(command))

def img_run(name, force, exercise_name = None, flag_name = None):

    image_settings = imgsettings.settings.get('images', {}).get(name)

    if not image_settings:
        raise log.FatalMsg("Image %s does not exist, run 'sfsdk img-add' to add it" % name)

    container_settings = imgsettings.get_container_conf(name)
    container_name = container_settings['name']

    container_exists = containers.container_exists(container_name)
    container_status = containers.container_status(container_name)

    if container_exists and not force:
        if container_status == 'running':
            raise log.FatalMsg("Container %s is already running, you can stop and rebuild it." % container_name)
        else:
            raise log.FatalMsg("Container %s is not running, run 'sfsdk img-run %s --force' to delete and re-run" % (container_name, name))

    data = [
            [ "Container name:", container_name ],
            [ "RDP port:", '127.0.0.1:%s' % image_settings['container']['external_rdp_port'] ],
            [ "RDP credentials:", 'sf:' + image_settings['container']['rdp_password']  ],
            [ "Application port:", '127.0.0.1:%s' % image_settings['container']['external_http_port'] ]
        ]

    if exercise_name:
        container_settings['environment']['EXERCISE'] = exrsettings.get_exercise_environment_variable(exercise_name)
        data.append(
            [ "Exercise name:", exercise_name ]
        )
    elif flag_name:
        container_settings['environment']['EXERCISE'] = exrsettings.generate_exercise_environment(flag_name)
        data.append(
            [ "Flag name:", flag_name ]
        )
    else:
        data.append(
            [ "Exercise name:", 'none' ]
        )        

    if container_status != None:
        containers.container_remove(container_name)

    containers.run_instance(container_settings)

    log.success("Container %s has been started" % container_name)
        

    log.tablify(data)


def img_stop(names):

    image_names = imgsettings.resolve_images(names)

    for name in image_names:

        image_settings = imgsettings.settings.get('images', {}).get(name)

        if not image_settings:
            raise log.FatalMsg("Image %s does not exist, run 'sfsdk img-add' to add it" % name)

        container_settings = imgsettings.get_container_conf(name)
        container_name = container_settings['name']

        container_was_running = (containers.container_status(container_name) != 'running')

        if not container_was_running:

            if not containers.container_remove(container_name):
                log.warn("Container %s is not running" % container_name)
            else:
                log.success("Container %s has been stopped" % container_name)

def img_watch(name, folder, count):

    image_settings = imgsettings.settings.get('images', {}).get(name)

    if not image_settings:
        raise log.FatalMsg("Image %s does not exist, run 'sfsdk img-add' to add it" % name)

    output_file = os.path.join(image_settings['build_dir'], '.sfsdk-watch.txt')
    
    container_settings = imgsettings.get_container_conf(name)

    container_name = container_settings['name']
    log.debug("Writing img-watch output to %s" % output_file)
    log.success("Please wait, preparing %s to run img-watch.." % (container_name))

    containers.watch_container(container_name, watch_folder = folder)
    last_json_output = []

    try:

        i = 0
        found = 0

        while True:

            if count == 0:
                pass
            elif i > count:
                break
            else:
                i+=1

            exit_code, output = containers.exec_on_container(
                    container_name,
                    'cat /tmp/sfsdkwatch.json',
                    )

            if exit_code == 0 and output:

                found += 1
                if found == 1:
                    log.success("Started watching changes in %s folder.." % (folder))


                try:
                    json_output = json.loads(output)
                except json.JSONDecodeError as e:
                    continue

                if sorted(json_output) != sorted(last_json_output):
                    last_json_output = json_output[:]

                    log.success("%s Changed files and folders:" % (time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())))

                    directory_list = '\n'.join(last_json_output)
                    log.block(directory_list, end='\n\n')

                    with open(output_file, 'w+') as stream:
                        stream.write(directory_list)

            time.sleep(1)

    except KeyboardInterrupt:
        pass

def img_ls():

    data = [ ]

    for image_name, image_data in imgsettings.settings['images'].items():

        img_running = (containers.container_status(image_data['container']['container_name']) == 'running')

        data.append([
            image_name,
            'Yes' if img_running else 'No',
            image_data['container']['container_name'] if img_running else '',
            image_data['container']['external_rdp_port'] if img_running else '',
            'sf:' + image_data['container']['rdp_password'] if img_running else '',
            image_data['container']['external_http_port'] if img_running else '',
            ]
        )

    log.tablify(
            data,
            header = [
                'Name',
                'Running',
                'Container name',
                'RDP port',
                'RDP creds',
                'Application port'
            ]
        )

def img_rm(names):

    image_names = imgsettings.resolve_images(names)

    for name in image_names:

        image_settings = imgsettings.settings.get('images', {}).get(name)

        if not image_settings:
            raise log.FatalMsg("Image %s does not exist, run 'sfsdk img-add' to add it" % name)

        try:
            imgsettings.remove_image(name)
        except log.FatalMsg as e:
            log.warn(str(e))
        else:
            log.success("Image %s data has been removed" % name)


def exercise_edit(source):

    authsettings.set_source(source)
    exrsettings.load(base_dir)

    _, management_url, username, password = authsettings.load_or_ask_for_creds()

    remote.login_to_hub(username, password, management_url)

    web.run_editor()


def config(source):

    log.warn("""Warning: this saves your credentials in %s in an insecure way. Skip this (press Ctrl-C) to be prompted when authenticating.""" % authsettings.settings_path)

    authsettings.set_source(source)

    source_type, management_url, username, password = authsettings.load_or_ask_for_creds(
            doing_config=True
        )

    if not source in authsettings.settings['sources']:
        authsettings.settings['sources'][source] = {}

    authsettings.settings['sources'][source]['type'] = source_type
    authsettings.settings['sources'][source]['username'] = username
    authsettings.settings['sources'][source]['password'] = password

    authsettings.save_to_file()

def publish(image_name, exercise_name, source):

    authsettings.set_source(source)

    source_type, management_url, username, password = authsettings.load_or_ask_for_creds()
    
    if source_type != 'hub':
        raise log.FatalMsg(
            "Error, a valid SecureFlag Exercise Hub account is required"
            )

    remote.login_to_hub(username, password, management_url)
    aws_creds = remote.get_s3_creds_via_hub()
    aws_arn = remote.get_arn_via_hub()

    tmp_dir_path = tempfile.mkdtemp(prefix='sfsdk-publish-archive-dir-')

    tmp_archive = tempfile.NamedTemporaryFile(
            prefix = 'sfsdk-publish-archive',
            )
    tmp_archive.close()

    log.debug('Building publish data in {tmp_dir_path} zipped to {tmp_archive.name}')

    shutil.copytree(
        imgsettings.settings['images'][image_name]['build_dir'], 
        os.path.join(tmp_dir_path, 'build')
    )

    shutil.copytree(
        os.path.join(exrsettings.base_dir, exercise_name), 
        os.path.join(tmp_dir_path, os.path.basename(exercise_name))
    )

    shutil.make_archive(tmp_archive.name, 'zip', tmp_dir_path)

    remote.push_to_s3(
            tmp_archive.name,
            aws_creds,
            username,
            aws_arn
    )

def exercise_ls(source):

    authsettings.set_source(source)

    log.info('Local exercises list\n' + '\n'.join(exrsettings.ls_exercises()))

def exercise_rm(names):

    for name in names:
        try:
            exrsettings.remove_exercise(name)
        except log.FatalMsg as e:
            log.warn(str(e))

def _get_argparser():

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='subparser')

    parser_add = subparsers.add_parser('img-add', help='Add new image')
    parser_add.add_argument(
        '-b', '--build-dir', dest='build_dir', help='Build folder')
    parser_add.add_argument(
        '-i', '--import-build', dest='from_dir', help='Import build directory')
    parser_add.add_argument('name', help='New image name')

    parser_ls = subparsers.add_parser('img-ls', help='List images')

    parser_build = subparsers.add_parser('img-build', help='Build an image')
    parser_build.add_argument('name', help='Image to build')
    parser_build.add_argument(
            '--source', dest='source', help='Source name')

    parser_run = subparsers.add_parser('img-run', help='Run last build')
    parser_run.add_argument('name', help='Project to run')
    parser_run.add_argument('-f', '--force', dest='force', help='Remove previous containers', action='store_true')
    parser_run.add_argument(
        '--exercise-name', '-en', dest='exercise_name', help='Pass exercise name'
        )
    parser_run.add_argument(
        '--flag-name', '-fn', dest='flag_name', help='Pass flag name'
        )
        
    parser_stop = subparsers.add_parser('img-stop', help='Stop running image')
    parser_stop.add_argument('names', nargs='+', help='Images names (or "all")')

    parser_shell = subparsers.add_parser('img-shell', help='Interact with a running image')
    parser_shell.add_argument('name', help='Run interactive shell on a running image')
    parser_shell.add_argument('-c', '--command', help='Command to execute', default='/bin/bash')

    parser_watch = subparsers.add_parser('img-watch', help='Keep track of the fs changes on a running image')
    parser_watch.add_argument(
        '-f', '--folder', dest='folder', help='Folder to watch', default='/home/sf')
    parser_watch.add_argument(
        '-c', '--count', dest='count', help=argparse.SUPPRESS, default=0, type=int)
    parser_watch.add_argument('name', help='Running image to watch')

    snapshot = subparsers.add_parser('img-snapshot', help='Save the current state as new')
    snapshot.add_argument(
        '--new-name', dest='new_name', help='The new image name')
    snapshot.add_argument(
        '-f', '--force', dest='force', help='Overwrite the export directory', action='store_true')
    snapshot.add_argument('name', help='Running image to snapshot')

    parser_rm = subparsers.add_parser('img-rm', help='Remove local image')
    parser_rm.add_argument('names', nargs='+', help='Images names')

    parser_exercise_ls = subparsers.add_parser('exercise-ls', help='List local exercises')
    parser_exercise_ls.add_argument(
            '--source', dest='source', help='Source name')

    exercise_edit = subparsers.add_parser('exercise-edit', help='Manage exercises via web editor')
    exercise_edit.add_argument(
            '--source', dest='source', help='Source name')

    exercise_rm = subparsers.add_parser('exercise-rm', help='Remove local exercise')
    exercise_rm.add_argument('names', nargs='+', help='Exercises names')

    publish = subparsers.add_parser('publish', help='Publish exercise and image on the SecureFlag Hub')
    publish.add_argument(
        '--image-name', dest='image_name', help='The image to publish', required=True)
    publish.add_argument(
        '--exercise-name', dest='exercise_name', help='The exercise name', required=True
        )
    publish.add_argument(
        '--source', dest='source', help='Source name')


    config = subparsers.add_parser('config', help='Configure authentication')
    config.add_argument(
            '--source', dest='source', help='Source name', default = 'default')

    parser.add_argument(
        '--verbose', '-v', dest='verbose', help='Verbose', action='store_true')

    return parser

def cli():

    parser = _get_argparser()

    kwargs = vars(parser.parse_args())

    if not kwargs.get('subparser'):
        parser.print_help()
        raise log.FatalMsg()

    imgsettings.load(base_dir)

    authsettings.load(base_dir)
    authsettings.set_source(kwargs.get('source'))

    # Initialize docker client
    containers.load()

    if not kwargs['subparser'].startswith('img-'):

        exrsettings.load(base_dir)
        kbsettings.load(base_dir)
        stacksettings.load(base_dir)
        frameworksettings.load(base_dir)

    # Set verbose and remove from arguments
    if kwargs['verbose']:
        log.set_verbose()

    del kwargs['verbose']

    globals()[kwargs.pop('subparser').replace('-', '_')](**kwargs)



if __name__ == "__main__":

    try:
        cli()
    except log.FatalMsg as e:
        msg = str(e)
        if msg: log.warn(str(e))
        sys.exit(1)
    except KeyboardInterrupt as e:
        print('')



# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from flask import Flask, request, jsonify, render_template, send_from_directory
from sfsdk import exrsettings, remote, stacksettings, kbsettings, frameworksettings
import random
import uuid
import os

ctoken = ''.join([ str(random.randint(0, 9)) for _ in range(19) ])

class FlaskWithDifferentDelimeters(Flask):
    jinja_options = Flask.jinja_options.copy()
    jinja_options.update(dict(
        block_start_string='<%',
        block_end_string='%>',
        variable_start_string='%%',
        variable_end_string='%%',
        comment_start_string='<#',
        comment_end_string='#>',
    ))

app = FlaskWithDifferentDelimeters(__name__)

@app.route('/')
def index():
    return render_template('index.html')

@app.route('/static/<path:path>')
def send_static(path): 
    return send_from_directory('static', path)

@app.route("/handler", methods = [ "POST" ])
def handler():

    req_data = request.get_json()
    
    action = req_data.get('action')
    if not action:
        return jsonify({'error': 'action required'})

    if action == 'getUserCToken':
        return jsonify({ 'ctoken': ctoken })

    # Anti-csrf token
    token = req_data.get('ctoken')
    if token != ctoken:
        return jsonify({ 'result': 'error', 'msg': 'Missing or wrong ctoken' })

    if action == 'getExercises':
        # Add local names
        return jsonify(
            [
                {**{ 'localName': n }, **e } for n, e in exrsettings.get_exercises().items() 
            ]
        )
    elif action == 'getExerciseDetails':
        # Add local names
        req_uuid = req_data.get('uuid') 
        exercise_local_name, exercise_data,  = exrsettings.get_exercise_name_data_by_uuid(req_uuid)
        exercise_data['localName'] = exercise_local_name
        return jsonify(exercise_data)
    elif action == 'addExercise':
        del req_data['ctoken']
        del req_data['action']

        req_data['uuid'] = f'sfsdk-{str(uuid.uuid4())}'
        exrsettings.save_exercise_meta(req_data)

        return jsonify({
            'result': 'success',
            'uuid': req_data['uuid']
        })
    elif action == 'updateExercise':
        req_uuid = req_data.get('uuid') 

        # Make sure uuid exists already
        _, old_exercise_data = exrsettings.get_exercise_name_data_by_uuid(req_uuid)
        if not old_exercise_data:
            return jsonify({
                'result': 'error',
                'uuid': 'unknown uuid'
            })

        del req_data['ctoken']
        del req_data['action']

        old_exercise_dir = os.path.join(
            exrsettings.workspace_dir,
            exrsettings.get_standard_exercise_name(old_exercise_data)
        )

        exrsettings.update_exercise_folder(old_exercise_dir, req_data)

        return jsonify({
            'result': 'success',
            'uuid': req_data['uuid']
        })
    elif action == "addTechnology":

        req_data['obj']['uuid'] = f'sfsdk-{str(uuid.uuid4())}'

        stack_folder = stacksettings.save_stack_meta(req_data['obj'])

        print(f"Saved to {stack_folder}")
        
        return jsonify({
            'result': 'success'
        })
    elif action == "updateTechnology":

        req_uuid = req_data["obj"].get('uuid') 

        # Make sure uuid exists already
        old_stack_name, old_stack_data = stacksettings.get_stack_name_data_by_uuid(req_uuid)
        if not old_stack_data:
            return jsonify({
                'result': 'error',
                'uuid': 'unknown uuid'
            })

        old_stack_dir = os.path.join(
            stacksettings.workspace_dir,
            old_stack_name
        )

        stacksettings.update_stack_folder(old_stack_dir, req_data["obj"])

        return jsonify({
            'result': 'success',
            'uuid': req_uuid
        })
    elif action == "deleteTechnology":
    
        req_uuid = req_data.get('uuid')

        stacksettings.delete_stack_by_uuid(req_uuid)

        return jsonify({
            'result': 'success'
        })

    elif action == "addVulnerability":
    
        req_data['obj']['uuid'] = f'sfsdk-{str(uuid.uuid4())}'

        kb_folder = kbsettings.save_kb_meta(req_data['obj'])

        print(f"Saved to {kb_folder}")

        return jsonify({
            'result': 'success'
        })
    elif action == "updateVulnerability":

        req_uuid = req_data["obj"].get('uuid') 

        # Make sure uuid exists already
        old_kb_name, old_kb_data = kbsettings.get_kb_name_data_by_uuid(req_uuid)
        if not old_kb_data:
            return jsonify({
                'result': 'error',
                'uuid': 'unknown uuid'
            })

        old_kb_dir = os.path.join(
            kbsettings.workspace_dir,
            old_kb_name
        )

        kbsettings.update_kb_folder(old_kb_dir, req_data["obj"])

        return jsonify({
            'result': 'success',
            'uuid': req_uuid
        })
    elif action == "deleteVulnerability":
        
        req_uuid = req_data.get('uuid')

        kbsettings.delete_kb_by_uuid(req_uuid)

        return jsonify({
            'result': 'success'
        })

    elif action in ("addFramework", "updateFramework"):
        frameworksettings.save_framework(req_data['name'])
        return jsonify({"result": "success"})
    elif action == 'removeFramework':
        frameworksettings.delete_framework_by_name(req_data['name'])
        return jsonify({"result": "success"})
    elif action == 'getAllKbs':
        local_kbs = [ k for k in kbsettings.get_kbs().values() ]
        return jsonify(remote.get_vulnerabilities_via_hub() + local_kbs)
    elif action == 'getFrameworks':
        local_frameworks = frameworksettings.get_frameworks()
        return jsonify(remote.get_frameworks_via_hub() + local_frameworks)
    elif action == 'getKBItem':
        kb_item = {}

        if req_data['uuid'] in [ k['uuid'] for k in kbsettings.get_kbs().values() ]:
            kb_name, kb_item = kbsettings.get_kb_name_data_by_uuid(req_data['uuid'])

        elif req_data['uuid'] in [ k['uuid'] for k in remote.get_vulnerabilities_via_hub() ]:
            kb_item = remote.get_vulnerability_via_hub_by_uuid(req_data['uuid'])

        return jsonify(kb_item)
    elif action == 'getAllStacks':
        local_stacks = [ s for s in stacksettings.get_stacks().values() ]
        return jsonify(remote.get_technologies_via_hub() + local_stacks)
    elif action == 'getStackItem':

        stack_item = {}

        if req_data['uuid'] in [ k['uuid'] for k in stacksettings.get_stacks().values() ]:
            _, stack_item = stacksettings.get_stack_name_data_by_uuid(req_data['uuid'])
            
        elif req_data['uuid'] in [ k['uuid'] for k in remote.get_technologies_via_hub() ]:
            stack_item = remote.get_technology_via_hub_by_uuid(req_data['uuid'])

        return jsonify(stack_item)

    else:
        return jsonify({ 'result': 'error', 'message': 'unknown action'})

def run_editor():

    app.run(host= '127.0.0.1')



# This file is part of the SecureFlag Platform.
# Copyright (c) 2022 SecureFlag Limited.

# This program is free software: you can redistribute it and/or modify  
# it under the terms of the GNU General Public License as published by  
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.

# You should have received a copy of the GNU General Public License 
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from sfsdk import log, utils
import docker
import time
import tempfile
import tarfile
import os
import json
import io
import boto3
import base64

import requests

dockerc = None

dockerfile_data = """
FROM secureflag/sf-community

COPY fs/ /
RUN chown -R sf:sf /home/sf
""".lstrip()

def load():

    global dockerc

    dockerc = docker.from_env()

    try:
        dockerc.info()
    except Exception as e:
        log.debug(str(e))
        raise log.FatalMsg("Error, can't communicate with the Docker daemon")

def image_exists(image_name):

    try:
        dockerc.images.get(image_name)
    except docker.errors.ImageNotFound as _:
        return False
    else:
        return True

def tag_image(image_name, tag):

    dockerc.images.get(image_name).tag(tag)

def run_instance(conf):

    global containers

    conf['detach'] = True
    try:
        dockerc.containers.run(**conf)
    except docker.errors.ImageNotFound as e:
        raise log.FatalMsg("Image has not been built yet, run 'sfsdk img-build %s' and retry" % conf['image'])
    except docker.errors.ContainerError as e:
        raise log.FatalMsg('A container named %s already exists' % conf['name'])
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

def containers_list():

    try:
         return [ c.name for c in dockerc.containers.list() ]
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

def container_exists(container_name):

    try:
        dockerc.containers.get(container_name)
    except (docker.errors.NotFound, requests.exceptions.ChunkedEncodingError) as e:
        return False
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

    return True


def container_status(container_name):

    try:
        return dockerc.containers.get(container_name).status
    except (docker.errors.NotFound, requests.exceptions.ChunkedEncodingError) as e:
        pass
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

def container_remove(container_name):

    try:
        container = dockerc.containers.get(container_name)
        container.stop(timeout=20)
        container.remove(force=True)
    except (docker.errors.NotFound, requests.exceptions.ChunkedEncodingError) as e:
        return False
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

    return True

def wait_until_running(container_name, delay = 1, count = 10):

    for i in range(count):

        try:
            container = dockerc.containers.get(container_name)
        except (docker.errors.NotFound, requests.exceptions.ChunkedEncodingError) as e:
            pass
        except docker.errors.APIError as e:
            raise log.FatalMsg(e.explanation)

        if container.status == 'running':
            return

        time.sleep(1)

    raise log.FatalMsg("Container failed running in %d seconds" % (delay*count))

def upload_paths(container_name, paths, dest):

    try:

        container = dockerc.containers.get(container_name)
        tar_file = tar_from_paths(paths)
        container.put_archive(dest, tar_file.read())

    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

def exec_on_container(container_name, cmd, tty = True, stream = False):

    try:
        container = dockerc.containers.get(container_name)

        if not stream:

            exit_code, output = container.exec_run(
                    cmd,
                    stdout = True,
                    tty = tty
                    )

            return exit_code, output.decode("utf-8", errors="ignore").rstrip("\r\n")

        else:

            result = container.exec_run(
                    cmd,
                    stdout = True,
                    tty = tty,
                    stream = True
                    )
            for line in result.output:
                log.block(line.decode("utf-8"))

    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)


def watch_container(container_name, watch_folder = '/home/sf', status_json = '/tmp/sfsdkwatch.json'):
            
    try:
        container = dockerc.containers.get(container_name)

        if container.exec_run(
                "ls /tmp/sfsdkwatch.py"
                ).exit_code != 0:
            upload_paths(container_name,
                    [ os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "sfsdkwatch.py"
                    ) ], '/tmp')

        _, pgrep_output = container.exec_run(
                "pgrep -f '/usr/bin/python3 /tmp/sfsdkwatch.py'",
                stdout = True
                )

        if not pgrep_output:
            if container.exec_run(
                    "/usr/bin/python3 -c 'import watchdog'"
                    ).exit_code != 0:
                if container.exec_run("sh -c \
                        'apt-get update && \
                        apt-get install -y python3-watchdog'",
                        environment = [ "DEBIAN_FRONTEND=noninteractive" ]).exit_code != 0:
                    raise log.FatalMsg("Error installing python-watchdog, check if internet is enabled")

            container.exec_run(
                "/usr/bin/python3 /tmp/sfsdkwatch.py '%s' '%s'" % (watch_folder, status_json), 
                detach = True
                )
    except docker.errors.NotFound:
        raise log.FatalMsg("Container is not running, start it with 'sfsdk run <image-name>' and retry")
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

def get_watched_paths(container_name):

    try:
        container = dockerc.containers.get(container_name)

        if container.exec_run(
                "ls /tmp/sfsdkwatch.json"
                ).exit_code != 0:
            return {}
        
        watch_data = container.exec_run("cat /tmp/sfsdkwatch.json").output
        
        if not watch_data:
            return {}
        else:
            return json.loads(watch_data)
    except ValueError as e:
        raise log.FatalMsg('Error reading the modified files')
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

def tar_from_paths(paths):
    f = tempfile.NamedTemporaryFile(delete=False)
    t = tarfile.open(mode='w', fileobj=f)

    for path in paths:
        abs_path = os.path.abspath(path)
        t.add(abs_path, arcname=os.path.basename(path), recursive=True)

    t.close()
    f.seek(0)
    return f

def tar_from_bytes(data):
    f = io.BytesIO()

    for chunk in data:
        f.write(chunk)

    f.seek(0)

    return tarfile.open(fileobj=f)

def is_empty_folder(snapshot_folder):

    if not os.path.isdir(snapshot_folder):
        os.makedirs(snapshot_folder)
        
    if [ x for x in os.listdir(snapshot_folder) if x != '.keep' ]:
        return False

    return True

def generate_dockerfile(build_folder, modified_paths = []):

    with open(
            os.path.join(build_folder, 'Dockerfile'),
            'w'
            ) as dockerfile:

        dockerfile.write(dockerfile_data)

def create_snapshot(snapshot_folder, snapshot_tmp, modified_paths):

    dest_home = os.path.join(snapshot_folder, 'fs/home/sf')
    
    utils.merge_tree(
        os.path.join(snapshot_tmp, 'sf'), 
        dest_home,
        preserve_symlinks = True,
    )

def extract_modified_from_tar(tar_file, modified_paths):

    snapshot_tmp = tempfile.mkdtemp(prefix='sdk.snapshot.')

    members_to_extract = []
    for m in tar_file.getmembers():
        file_path = os.path.join('/home', m.name)

        # Standardize output using an ending / if folders.
        if m.isdir():
            file_path = os.path.join(file_path, '')

        # Extract everything starts with modified_paths
        if any((mp for mp in modified_paths + [ '/home/sf/exercise' ] if file_path.startswith(mp) and not m in [n.name for n in members_to_extract])):
            members_to_extract.append(m)

    tar_file.extractall(path=snapshot_tmp, members=members_to_extract)

    return snapshot_tmp

def save_container_snapshot(container_name, snapshot_folder, modified_paths):

    try:
        container = dockerc.containers.get(container_name)
    except (docker.errors.NotFound, requests.exceptions.ChunkedEncodingError) as e:
        raise log.FatalMsg("Can't find container %s" % container_name)
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

    try:
        tar_data, _ = container.get_archive('/home/sf')
    except Exception as e:
        log.debug(str(e))
        raise log.FatalMsg("Error pulling data from container")

    tar_file = tar_from_bytes(tar_data)

    try:
        snapshot_tmp = extract_modified_from_tar(
            tar_file, 
            modified_paths
        )
    except Exception as e:
        log.debug(str(e))
        raise log.FatalMsg("Error extracting data from container")

    create_snapshot(snapshot_folder, snapshot_tmp, modified_paths)

def build_image(exercise_path, image_name):

    try:
        return dockerc.api.build(
                path=exercise_path,
                tag=image_name,
                decode=True
                )
    except docker.errors.APIError as e:
        log.warn("Error building %s from %s" % (image_name, exercise_path))
        raise log.FatalMsg(e.explanation)

def image_remove(image_name):

    try:
        dockerc.images.remove(image_name, force=True)
    except (docker.errors.NotFound, requests.exceptions.ChunkedEncodingError) as e:
        return False
    except docker.errors.APIError as e:
        raise log.FatalMsg(e.explanation)

    return True